;;; Dezyne --- Dezyne command line tools
;;;
;;; Copyright © 2015, 2016, 2017, 2019, 2020, 2021, 2022, 2023 Janneke Nieuwenhuizen <janneke@gnu.org>
;;; Copyright © 2019 Rob Wieringa <rma.wieringa@gmail.com>
;;; Copyright © 2017, 2018 Rutger van Beusekom <rutger@dezyne.org>
;;; Copyright © 2024 Paul Hoogendijk <paul@dezyne.org>
;;;
;;; This file is part of Dezyne.
;;;
;;; Dezyne is free software: you can redistribute it and/or modify it
;;; under the terms of the GNU Affero General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Dezyne is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Affero General Public License for more details.
;;;
;;; You should have received a copy of the GNU Affero General Public
;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.
;;;
;;; Commentary:
;;;
;;; Code:

(define-module (dzn code language python)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)

  #:use-module (scmackerel indent)

  #:use-module (dzn ast ast)
  #:use-module (dzn ast normalize)
  #:use-module (dzn ast)
  #:use-module (dzn code)
  #:use-module (dzn code scmackerel python)
  #:use-module (dzn code util)
  #:use-module (dzn misc)
  #:export (<capture-variable>
            python:capture-name
            python:defer-equality*
            python:defer-variable*
            python:number-argument
            python:out-ref))

;;;
;;; Ast.
;;;
(define-ast <capture-variable> (<variable>)
  (depth))


;;;
;;; Accessors.
;;;
(define-method (python:out-ref (o <formal>) (name <string>))
  (cond ((ast:in? o) name)
        ((ast:out? o) (string-append "out " name))
        ((ast:inout? o) (string-append "ref " name))))

(define-method (python:out-ref (o <formal>) (name <top>))
  (python:out-ref o))

(define-method (python:out-ref (o <formal>) (name <number>))
  (python:out-ref o (number->string name)))

(define-method (python:number-argument (o <formal>))
  (code:number-argument o))

(define-method (python:out-ref (o <formal>))
  (python:out-ref o (.name o)))

(define-method (python:defer-variable* (o <defer>))
  (let* ((variables (ast:defer-variable* o))
         (depth (length (filter (is? <defer>) (tree:path o)))))
    (map (compose
          (cute graft (tree:parent o) <>)
          (cute make <capture-variable> #:name <> #:type.name <> #:depth depth))
         (map .name variables)
         (map .type.name variables))))

(define-method (python:defer-equality* (o <defer>))
  (filter (compose not (is? <extern>) .type) (python:defer-variable* o)))

(define-method (python:capture-name (o <variable>))
  (code:capture-name o))

(define-method (python:capture-name (o <capture-variable>))
  (simple-format #f "~a~a" (code:capture-name o) (.depth o)))

(define-method (python:model (o <root>))
  (let* ((models (ast:model** o))
         (models (filter (negate (disjoin (is? <type>)
                                          (is? <namespace>)
                                          ast:imported?))
                         models))
         (models (ast:topological-model-sort models))
         (models (map code:annotate-shells models)))
    models))


;;;
;;; Normalizations.
;;;
(define (python:normalize ast)
  (parameterize ((%normalize:short-circuit? code:short-circuit?))
    ((compose
      (with-root code:annotate-shells)
      (with-root add-reply-port)
      (with-root normalize:event+illegals)
      (with-root remove-otherwise)
      (with-root code:add-calling-context))
     ast)))


;;;
;;; Entry point.
;;;
(define* (ast-> root #:key (dir ".") model)
  "Entry point."

  (code:foreign-conflict? root)

  (parameterize ((%language "python")
                 (%member-prefix "self.")
                 (%name-infix ".")
                 (%type-infix ".")
                 (%type-prefix "dzn."))
    (let ((root (python:normalize root)))
      (let* ((generator (sm:indenter (cute print-code-ast root) #:width 4 #:gnu? #f))
             (generator (cute display (string-delete (string->char-set "{}") (warn 'org (with-output-to-string generator)))))
             (file-name (code:root-file-name root dir ".py")))
        (code:dump generator #:file-name file-name))

      (when model
        (let ((model (ast:get-model root model)))
          (when (is-a? model <component-model>)
            (let* ((generator (sm:indenter (cute print-main-ast model) #:width 4 #:gnu? #f))
                   (generator (cute display (string-delete (string->char-set "{}") (with-output-to-string generator))))
                   (file-name (code:source-file-name "main" dir ".py")))
              (code:dump generator #:file-name file-name))))))))
