;;; Dezyne --- Dezyne command line tools
;;;
;;; Copyright © 2022, 2023, 2024 Janneke Nieuwenhuizen <janneke@gnu.org>
;;; Copyright © 2024 Paul Hoogendijk <paul@dezyne.org>
;;;
;;; This file is part of Dezyne.
;;;
;;; Dezyne is free software: you can redistribute it and/or modify it
;;; under the terms of the GNU Affero General Public License as
;;; published by the Free Software Foundation, either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; Dezyne is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Affero General Public License for more details.
;;;
;;; You should have received a copy of the GNU Affero General Public
;;; License along with Dezyne.  If not, see <http://www.gnu.org/licenses/>.

(define-module (dzn code scmackerel python)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-71)

  #:use-module (ice-9 match)
  #:use-module (ice-9 rdelim)
  #:use-module (scmackerel python)

  #:use-module (dzn ast ast)
  #:use-module (dzn ast)
  #:use-module (dzn code)
  #:use-module (dzn code ast)
  #:use-module (dzn code shared)
  #:use-module (dzn code scmackerel code)
  #:use-module (dzn code language dzn)
  #:use-module (dzn code language python)
  #:use-module (dzn config)
  #:use-module (dzn misc)
  #:export (print-code-ast
            print-main-ast))

(define make-sm:cpp (@@(scmackerel cpp) make-sm:cpp))


;;;
;;; Helpers.
;;;
(define (file-comments file-name)
  (let ((dir (string-append %template-dir "/python")))
    (list
     (sm:comment* (code:caption file-name #:line-comment #\#))
     (sm:comment* "\n")
     (sm:comment*
      (with-input-from-file (string-append dir "/" file-name)
        read-string)))))

(define (python:binding->connect binding)
  (let* ((provides
          requires
          (code:provides+requires-end-point binding))
         (port (code:type-name (.type (.port provides)))))
    (sm:call (name "runtime.connect")
             (arguments
              (list (sm:member* (code:end-point->string provides))
                    (sm:member* (code:end-point->string requires)))))))

(define-method (python:sm:statement* (o <compound>))
  (ast:statement* o))

;; XXX FIXME: introduce <block> statement instead of
;; <blocking-compound>, like VM?
(define-method (python:sm:statement* (o <blocking-compound>))
  (let* ((port-name (.port.name o))
         (block (sm:call (name "dzn.pump.port_block")
                         (arguments
                          (list "dzn_locator"
                                "this"
                                (string-append
                                 (%member-prefix)
                                 port-name))))))
    `(,@(ast:statement* o)
      ,block)))

(define (python:sm:formal->temp formal)
  (let ((name (.name formal))
        (type (code:type-name formal)))
    (sm:variable (type type)
                 (expression (if (ast:inout? formal) name
                                 (sm:call (name "default")
                                          (arguments (list type)))))
                 (name (string-append "dzn_" name)))))

(define (python:temp->formal formal)
  (let ((name (.name formal)))
    (sm:assign* name (string-append "dzn_" name))))

(define-method (python:import->import (o <import>))
  (sm:cpp-sm:using* (string-append "dzn." (code:file-name o))))

(define-method (python:local-imports (o <root>))
  (let ((imports (tree:collect o (cute is-a? <> <import>))))
    (filter (negate ast:imported?) imports)))


;;;
;;; Ast->code.
;;;
(define-method (ast->code (o <guard>))
  (let ((expression (.expression o))
        (statement (.statement o)))
    (cond
     ((is-a? expression <otherwise>)
      (ast->code statement))
     ((ast:literal-true? expression)
      (cond
       ((and (is-a? statement <compound>)
             (as (python:sm:statement* statement) <pair>))
        =>
        (lambda (statements)
          (sm:statements* (map ast->code statements))))
       ((is-a? statement <compound>)
        (sm:statement*))
       (else
        (ast->code statement))))
     (else
      (sm:if* (ast->expression expression)
              (ast->code statement))))))

;;; imperative
(define-method (ast->code (o <blocking-compound>))
  (let ((statements (python:sm:statement* o)))
    (sm:compound* (map ast->code statements)))) ;;; XXX new sm:compound*

(define-method (ast->expression (o <top>) (sm:formal <formal>))
  (ast->expression o))

(define-method (ast->expression (name <string>) (formal <formal>))
  (python:out-ref formal name))

(define-method (ast->expression (o <reference>) (formal <formal>))
  (let* ((name (.name o))
         (name (if (ast:member? (.variable o)) (string-append "self." name)
                   name)))
    (ast->expression name formal)))

(define-method (ast->code (o <action>))
  (let* ((action-name (code:event-name o))
         (formals (ast:formal* (.event o)))
         (arguments (ast:argument* o))
         (arguments (map ast->expression arguments formals))
         (action (sm:call (name (string-append (%member-prefix) action-name))
                          (arguments arguments))))
    action))

(define-method (ast->code (o <call>))
  (let* ((formals (ast:formal* (.function o)))
         (arguments (ast:argument* o))
         (arguments (map ast->expression arguments formals)))
    (sm:call-method
     (name (.function.name o))
     (arguments arguments))))

(define-method (ast->code (o <reply>))
  (let ((p (tree:parent o)))
    (cond
     ((or (is-a? p <guard>) (is-a? p <if>))
      (ast->code (code:wrap-compound o)))
     (else
      (let* ((type (ast:type o))
             (reply-port (.port o))
             (port-name (.name reply-port))
             (out-binding (string-append (%member-prefix)
                                         (code:out-binding (.port o)))))
        (sm:statements*
         `(,@(if (is-a? type <void>) '()
                 `(,(sm:assign*
                     (sm:member* (%member-prefix) (code:reply-var type))
                     (ast->expression (.expression o)))))
           ,@(if (not (code:port-release? o)) '()
                 `(,(sm:call (name "dzn.pump.port_release")
                             (arguments
                              (list "dzn_locator"
                                    "this"
                                    (string-append
                                     (%member-prefix)
                                     (.port.name o))))))))))))))

(define-method (ast->code (o <illegal>))
  (sm:call (name (sm:member* (%member-prefix) "dzn_runtime.illegal"))))

(define-method (python:->formal (o <formal>))
  (let* ((type (code:type-name o))
         (type (python:out-ref o type)))
    (sm:formal (type type)
               (name (.name o)))))

;; FIXME: c&p from c++
(define-method (ast->code (o <defer>))
  (define (variable->defer-variable o)
    (sm:variable
     (type (code:type-name (ast:type o)))
     (name (python:capture-name o))
     (expression (code:member-name o))))
  (define (variable->equality o)
    (sm:equal* (ast->expression o)
               (python:capture-name o)))
  (let* ((variables (python:defer-variable* o))
         (locals (code:capture-local o))
         (equality (python:defer-equality* o))
         (condition (if (not (code:defer-condition o)) "true"
                        (sm:and*
                         (map variable->equality equality))))
         (statement (.statement o))
         (statement (if (is-a? statement <compound>) statement
                        (code:wrap-compound statement)))
         (statement (ast->code statement))
         (statement
          (sm:compound*
           `(,@(map variable->defer-variable variables)
             ,(sm:call (name "dzn.Runtime.defer")
                       (arguments
                        (list "this"
                              (sm:function
                               (captures
                                (cons* "this"
                                       (map python:capture-name variables)))
                               (statement
                                (sm:compound*
                                 (sm:return* condition))))
                              (sm:function
                               (captures
                                (cons* "&" (map .name locals)))
                               (statement statement)))))))))
    statement))

(define-method (python:->argument (o <formal>))
  (let* ((name (.name o))
         (name (if (ast:in? o) name
                   (string-append "dzn_" name))))
    (python:out-ref o name)))

(define-method (ast->code (o <out-bindings>))
  (define (formal->assign formal)
    (let* ((name (.name formal))
           (name (if (ast:in? formal) name
                     (string-append "dzn_" name))))
      (sm:assign* name
                  (sm:member* (%member-prefix) (.variable.name formal)))))
  (let ((formals (ast:formal* o)))
    (if (null? formals) (sm:statement*)
        (sm:assign*
         (sm:member* (%member-prefix) (code:out-binding (.port o)))
         (sm:function
          (statement
           (sm:compound*
            (map formal->assign formals))))))))


;;;
;;; Root.
;;;
(define-method (root->statements (o <root>))
  (let ((enums (code:enum* o)))
    (append
     ;;; FIXME: dollar support
     (map (compose
           (cute string-append <> "\n") .value)
          (ast:data* o))
     (append-map code:->namespace
                 enums
                 (map code:enum->sm:enum-struct enums)))))


;;;
;;; Interface.
;;;

;; C&$ from c++:model->statements
(define-method (interface->statements (o <interface>))
  (let* ((interface
          (sm:struct
            (name (dzn:model-name o))
            (types (map code:enum->sm:enum-struct (code:public-enum* o)))
            (methods
             (list
              (sm:constructor
               (struct #f)
               (formals (list "provides=runtime.Port ()" "requires=runtime.Port ()"))
               (statement
                (sm:statements*
                 `(,(sm:struct
                      (name "In")
                      (parents (list "runtime.Port"))
                      (methods
                       (list
                        (sm:constructor
                         (struct #f)
                         (formals (list "port"))
                         (statement
                          (sm:statements*
                           (append
                            (list "runtime.Port.__init__ (self, port.dzn_name, port.dzn_component)")
                            (map (lambda (e)
                                   (sm:assign* (string-append "self." (.name e))
                                               "None")) (ast:in-event* o)))))))))
                   ,(sm:assign*
                     "self.in_port"
                     "In (provides)")
                   ,(sm:struct
                      (name "Out")
                      (parents (list "runtime.Port"))
                      (methods
                       (list
                        (sm:constructor
                         (struct #f)
                         (formals (list "port"))
                         (statement
                          (sm:statements*
                           (append
                            (list "runtime.Port.__init__ (self, port.dzn_name, port.dzn_component)")
                            (map (lambda (e)
                                   (sm:assign*
                                    (string-append (%member-prefix) (.name e))
                                    "None"))
                                 (ast:out-event* o)))))))))
                   ,(sm:assign*
                     "self.out_port"
                     "Out (requires)"))))))))))
    (code:->namespace o interface)))


;;;
;;; Component.
;;;

;; C&P from c++:component-model->statements
(define-method (component-model->statements (o <component-model>))
  (define (member->assign o)
    (sm:assign* (sm:member* (%member-prefix) (.name o))
                (ast->expression (.expression o))))
  (define (reply-type->assign o)
    (sm:assign* (sm:member* (code:reply-var o)) "None"))
  (define (trigger->method component trigger)
    (let* ((trigger
            statement
            (match o
              ;;              (($ <foreign>)
              ;;               (values trigger #f))
              (($ <component>)
               (let* ((on (tree:ancestor trigger <on>))
                      (statement (.statement on)))
                 (values trigger (ast->code statement))))))
           (formals (ast:formal* trigger))
           (formals (map python:->formal formals))
           (type (ast:type trigger))
           (typed? (not (is-a? type <void>)))
           (statement (if typed?
                          (sm:compound*
                           (append
                            (sm:compound-statements statement)
                            `(,(sm:return* (sm:member* (code:reply-var type))))))
                          statement)))
      (sm:method (struct component)
                 (type (if (not (is-a? o <foreign>)) "void"
                           (string-append "virtual " (code:type-name type))))
                 (name (code:event-slot-name trigger))
                 (formals formals)
                 (statement statement))))
  (define (function->method component function)
    (let* ((type (code:type-name (ast:type function)))
           (name (.name function))
           (formals (ast:formal* function))
           (formals (map python:->formal formals))
           (statement (.statement function))
           (statement (ast->code statement)))
      (sm:method (struct component)
                 (type type)
                 (name name)
                 (formals formals)
                 (statement statement))))
  (define (provides->assign port)
    (let* ((port-name (.name port))
           (type (.type port)))
      (sm:assign*
       (sm:member* port-name)
       (sm:call
        (name (code:type-name type))
        (arguments
         (list (sm:assign*
                "provides"
                (sm:call (name "runtime.Port")
                         (arguments (list (string-append "'" port-name "'")
                                          "self"))))))))))
  (define (requires->assign port)
    (let* ((port-name (.name port))
           (type (.type port)))
      (sm:assign*
       (sm:member* port-name)
       (sm:call
        (name (string-append "dzn." (string-join (ast:full-name type) ".")))
        (arguments
         (list (sm:assign*
                "requires"
                (sm:call (name "runtime.Port")
                         (arguments (list (string-append "'" port-name "'")
                                          "self"))))))))))
  (define injected->assign requires->assign)
  (define (trigger->event-slot trigger)
    (let* ((port (.port trigger))
           (port-name (.name port))
           (event (.event trigger))
           (type (ast:type trigger))
           (type-name (code:type-name type))
           ;;           (type-name (string-join (ast:full-name type) "."))
           (typed? (not (is-a? type <void>)))
           (direction (code:direction event))
           (event-name-string (simple-format #f "~s" (.name event))))
      (sm:assign*
       (sm:member* (%member-prefix) (code:event-name trigger))
       (sm:function
        (formals (list "*args"))
        (statement
         (sm:call
          (name (string-append "runtime." (if typed? "r" "") "call_" direction))
          (arguments
           (list
            "self"
            (string-append "lambda: self." (code:event-slot-name trigger) " (*args)")
            (if typed?
                (cond
                 ((is-a? type <bool>)
                  (list (sm:member* port-name) "," event-name-string ","
                        "lambda val: runtime.bool_to_string (val)"))
                 ((is-a? type <int>)
                  (list (sm:member* port-name) "," event-name-string
                        "," "lambda val: runtime.int_to_string (val)"))
                 ((is-a? type <enum>)
                  (list (sm:member* port-name) "," event-name-string
                        ","
                        (string-append "lambda val: " type-name
                                       ".to_string[val]"))))
                (list (sm:member* port-name) "," event-name-string))))))))))
  (let* ((enums (filter (is? <enum>) (code:enum* o)))
         (component
          (sm:struct
            (name (dzn:model-name o))
            (parents '("runtime.Component"))
            (types (map code:enum->sm:enum-struct enums))
            (members '())))
         (component
          (sm:struct
            (inherit component)
            (methods
             `(,(sm:constructor
                 (struct component)
                 (formals (list "locator" "name=''" "parent=None"))
                 (statement
                  (sm:compound*
                   `("runtime.Component.__init__ (self, locator, name, parent)"
                     "locator.get (runtime.Runtime).flushes (self)"
                     ,@(map reply-type->assign (code:reply-types o))
                     ,@(map member->assign (ast:member* o))
                     ,@(map provides->assign (ast:provides-port* o))
                     ,@(map requires->assign (ast:requires-port* o))
                     ,@(map trigger->event-slot (ast:in-triggers o))))))
               ,@(map (cute trigger->method component <>) (ast:in-triggers o))
               ,@(map (cute function->method component <>) (ast:function* o)))))))
    (code:->namespace o component)))

(define-method (interface->statements (o <component>))
  (component-model->statements o))


;;;
;;; Foreign.
;;;
(define-method (interface->statements (o <foreign>))
  '())


;;;
;;; System.
;;;

;;; C&P from c++:main-*
(define-method (system->statements (o <system>))
  (let* ((injected-instances (code:injected-instance* o))
         (injected? (pair? injected-instances)))
    (define (provides->assign port)
      (let* ((port-name (.name port))
             (type (.type port)))
        (sm:assign*
         (sm:member* port-name)
         (sm:call
          (name (code:type-name type))
          (arguments
           (list (sm:assign*
                  "provides"
                  (sm:call (name "runtime.Port")
                           (arguments (list (string-append "'" port-name "'")
                                            "self"))))))))))
    (define (requires->assign port)
      (let* ((port-name (.name port))
             (type (.type port)))
        (sm:assign*
         (sm:member* port-name)
         (sm:call
          (name (string-append "dzn." (string-join (ast:full-name type) ".")))
          (arguments
           (list (sm:assign*
                  "requires"
                  (sm:call (name "runtime.Port")
                           (arguments (list (string-append "'" port-name "'")
                                            "self"))))))))))
    (define (port->expression port)
      (let ((other-end (ast:other-end-point port)))
        (string-append (%member-prefix)
                       (.instance.name other-end)
                       "."
                       (.port.name other-end))))
    (define (port->member port)
      (let ((other-end (ast:other-end-point port)))
        (sm:variable
         (type (code:type-name (.type port)))
         (name (.name port))
         (expression (and (not injected?)
                          (port->expression port))))))
    (define (instance->expression instance)
      (let ((instance-name (.name instance))
            (type (code:type-name (.type instance))))
        (sm:call (name type)
                 (arguments
                  (list "locator"
                        (simple-format #f "name='~a'" instance-name)
                        "parent=self")))))
    (define (instance->assignment instance)
      (let ((name (.name instance)))
        (sm:assign* (sm:member* name)
                    (instance->expression instance))))
    (define (port->assignment port)
      (let ((port-name (.name port)))
        (sm:assign*
         (sm:member* port-name)
         (sm:call
          (name "runtime.mixin_meta")
          (arguments
           (list
            (port->expression port)
            (sm:member* port-name)))))))
    (let* ((instances (code:instance* o))
           (bindings (code:component-binding* o))
           (injected-bindings (code:injected-binding* o))
           (system
            (sm:struct
              (name (dzn:model-name o))
              (parents '("runtime.Component"))))
           (system
            (sm:struct
              (inherit system)
              (methods
               (list
                (sm:constructor
                 (struct system)
                 (formals (list "locator" "name=''" "parent=None"))
                 (statement
                  (sm:compound*
                   `("runtime.Component.__init__ (self, locator, name, parent)"
                     "locator.get (runtime.Runtime).flushes (self)"
                     ,@(map instance->assignment (warn 'ints instances))
                     ,@(map provides->assign (ast:provides-port* o))
                     ,@(map requires->assign (ast:requires-port* o))
                     ,@(map port->assignment (ast:provides-port* o))
                     ,@(map port->assignment (ast:requires-port* o))
                     ,@(map python:binding->connect bindings))))))))))
      (code:->namespace o system))))

(define-method (interface->statements (o <system>))
  (system->statements o))


;;;
;;; Main.
;;;

(define-method (main-getopt)
  (sm:function
   (type "static bool")
   (name "dzn_getopt")
   (formals (list (sm:formal (type "String[]") (name "args"))
                  (sm:formal (type "String") (name "option"))))
   (statement
    (sm:compound*
     (sm:return*
      (sm:call (name "Array.Exists")
               (arguments
                (list "args"
                      (sm:function (formals '("s"))
                                   (statement (sm:equal* "s" "option")))))))))))

(define (add-newline o)
  (string-append (sm:code->string o) "\n"))
(define-method (python:fill-event-map (o <component-model>))
  (define (trigger->event-map-entry trigger)
    (let* ((port (.port.name trigger))
           (event (.event.name trigger))
           (port-event (simple-format #f "~a.~a" port event))
           (formals (ast:formal* trigger))
           (formals (code:number-formals formals))
           (arguments (map python:number-argument formals))
           (e-port-event-string (simple-format #f "e['~a']" port-event)))
      (sm:statements*
       (list
        (sm:assign*
         e-port-event-string
         (sm:function
          (formals (list "*args"))
          (statement
           (sm:call
            (name (string-append "m." (code:event-name trigger)))
            (arguments arguments)))))))))
  (define (port->assing-name port)
    (let ((name (.name port)))
      (string-append
       "m." name ".out_port.dzn_component = c\n"
       "m." name ".out_port.dzn_name = '" name "'\n")))
  (define (port->add-logger port)
    (let ((name (.name port)))
      (string-append
       "m." name ".in_port.dzn_component = c\n"
       "m." name ".in_port.dzn_name = '" name "'\n"
       "def log_flush ():{\n"
       "    sys.stderr.write ('" name ".<flush>\\n')\n"
       "    m.dzn_runtime.flush (m." name ".in_port.dzn_component)}\n"
       "e['" name ".<flush>'] = log_flush\n")))
  (define (trigger->event-slot trigger)
    (let* ((type (ast:type trigger))
           (type-name (code:type-name type))
           (typed? (not (is-a? type <void>)))
           (bool? (is-a? type <bool>))
           (direction (code:direction trigger))
           (port-name (.port.name trigger))
           (event-name (.event.name trigger))
           (m-dot-port-name (simple-format #f "m.~a" port-name))
           (event-name-string (simple-format #f "'~a'"  event-name)))
      (add-newline
       (sm:assign*
        (string-append "m." (code:event-name trigger))
        (sm:function
         (formals (list "*args"))
         (statement
          (sm:call
           (name (if typed? "log_valued" (string-append "log_" direction)))
           (arguments
            `(
              ,m-dot-port-name
              ,event-name-string
              ,@(if (equal? direction "in") `("e") '())
              ,@(if typed?
                    (cond
                     ((is-a? type <bool>)
                      `("lambda event : runtime.event_to_bool (event)"))
                     ((is-a? type <int>)
                      `("lambda event : runtime.event_to_int (event)"))
                     ((is-a? type <enum>)
                      `(,(string-append "lambda event : runtime.event_to_enum ("
                                        type-name ", event)"))))
                    '()))))))))))
  (let ((name (.name o)))
    `(,(string-append "def " name "_fill_event_map (m):{\n")
      "c = runtime.Component (m.dzn_locator)\n"
      "c.dzn_name = '<external>'\n"
      "m.dzn_locator.get (runtime.Runtime).flushes (c, False)\n"
      "e = dict()\n"
      ,@(map trigger->event-map-entry (ast:provides-in-triggers o))
      ,@(map trigger->event-map-entry (ast:requires-out-triggers o))
      "\n"
      ,@(map port->assing-name (ast:provides-port* o))
      ,@(map port->add-logger (ast:requires-port* o))
      "\n"
      ,@(map trigger->event-slot (ast:provides-out-triggers o))
      ,@(map trigger->event-slot (ast:requires-in-triggers o))
      "return e\n\n}")))

(define-method (main-struct (o <component-model>))
  (let ((name (.name o)))
    (format #f "
def main ():{
    def illegal ():{
        sys.stderr.write('illegal')
        sys.exit (0)}
    loc = locator.Locator ()
    rt = runtime.Runtime (illegal)
    sut = dzn.~a (loc.set (rt), name='sut')

    event_map = ~a_fill_event_map (sut)

    s = sys.stdin.readline ().strip ()
    while s:{
        if s in event_map.keys ():{
            event_map[s] ()}
        s = sys.stdin.readline ().strip ()}}
"
            name name)))


;;;
;;; Entry points.
;;;
(define-method (print-code-ast (o <root>))
  (let* ((models (code:model* o))
         (comment (dzn:comment o))
         (comment (and
                   comment
                   (string-join (map (lambda (l) (if (equal? l "") l
                                                     (string-append "# " l)))
                                     (string-split comment #\newline))
                                "\n" 'suffix)))
         ;; (file-includes (append-map code:component-include* models))
         ;; (local-imports (python:local-imports o))
         (code (sm:code
                 (statements
                  `(,(code:generated-comment o #:line-comment "#")
                    ,@(if (not comment) '()
                          (list (sm:comment* comment)))
                    ,(sm:cpp-sm:using* "runtime")
                    ,(sm:cpp-sm:using* "dzn")
                    ;; ,@(map python:import->import local-imports)
                    ,@(root->statements o)
                    ,@(append-map interface->statements models)
                    ,(code:version-comment #:line-comment "#"))))))
    ((@@ (scmackerel python) set-record-printers!))
    (display (with-output-to-string (cute display code)))))

(define-method (print-main-ast (o <component-model>))
  (let ((code (sm:code
                (statements
                 `("#! /usr/bin/env python3\n"
                   ,(code:generated-comment o #:line-comment "#")
                   ,(sm:cpp-sm:using* "sys")
                   ,(sm:cpp-sm:using* "os")
                   "sys.path.insert (0, os.path.dirname (sys.argv[0]))\n"
                   "#\n"
                   ,(sm:cpp-sm:using* "dzn")
                   ,(sm:cpp-sm:using* (string-append "dzn." (code:file-name o)))
                   ,(sm:cpp-sm:using* "locator")
                   ,(sm:cpp-sm:using* "runtime")
                   "\n"
                   ,@(file-comments "main.py")
                   ,@(python:fill-event-map o)
                   ,(main-struct o)
                   "if __name__ == '__main__':{\n"
                   "    main ()}\n"
                   ,(code:version-comment #:line-comment "#"))))))
    ((@@ (scmackerel python) set-record-printers!))
    (display (with-output-to-string (cute display code)))))
